/*
  serial.C - UART functions 

  These functions provide interrupt driven or 
  non-interrupt driven serial communication.

*/

#include <msp430x16x.h>
#include "dz1611_config.h"
#include "serial_config.h"
#include "serial.h"

////////////////////////////////////////////////////////////////
//   UART0 variables
////////////////////////////////////////////////////////////////
#ifdef UART0_INT_DRIVEN
#ifdef USE_UART0
char uart0_rx_buf[UART0_RX_BUFSIZE];
char uart0_tx_buf[UART0_TX_BUFSIZE];

unsigned char uart0_rx_head, uart0_rx_tail;
unsigned char uart0_tx_head, uart0_tx_tail;
unsigned char uart0_flags;
#endif // USE_UART0
#endif // UART0_INT_DRIVEN

////////////////////////////////////////////////////////////////
//   UART1 variables
////////////////////////////////////////////////////////////////
#ifdef UART1_INT_DRIVEN
#ifdef USE_UART1
char uart1_rx_buf[UART1_RX_BUFSIZE];
char uart1_tx_buf[UART1_TX_BUFSIZE];

unsigned char uart1_rx_head, uart1_rx_tail;
unsigned char uart1_tx_head, uart1_tx_tail;
unsigned char uart1_flags;
#endif //USE_UART1
#endif // UART1_INT_DRIVEN


#ifdef UART0_INT_DRIVEN
    #ifdef USE_UART0
        char uart0_rx_buf[UART0_RX_BUFSIZE];
        char uart0_tx_buf[UART0_TX_BUFSIZE];

        unsigned char uart0_rx_head, uart0_rx_tail;
        unsigned char uart0_tx_head, uart0_tx_tail;
        unsigned char uart0_flags;

        #ifdef UART0_USING_CTS_RTS
            volatile unsigned char uart0_tx_enable;
        #endif // UART0_USING_CTS_RTS
    #endif // USE_UART0
#endif // UART0_INT_DRIVEN

////////////////////////////////////////////////////////////////
//   UART1 variables
////////////////////////////////////////////////////////////////
#ifdef UART1_INT_DRIVEN
    #ifdef USE_UART1
        char uart1_rx_buf[UART1_RX_BUFSIZE];
        char uart1_tx_buf[UART1_TX_BUFSIZE];

        unsigned char uart1_rx_head, uart1_rx_tail;
        unsigned char uart1_tx_head, uart1_tx_tail;
        unsigned char uart1_flags;

        #ifdef UART1_USING_CTS_RTS
            volatile unsigned char uart1_tx_enable;
            void PORT_ISR(void);
        #endif // UART1_USING_CTS_RTS
    #endif //USE_UART1
#endif // UART1_INT_DRIVEN


////////////////////////////////////////////////////////////////
//   UART0 functions
////////////////////////////////////////////////////////////////
#ifdef USE_UART0
    #ifdef UART0_INT_DRIVEN
        void UART0_ClearTxBuffer(void)
        {
            uart0_tx_head = 0;
            uart0_tx_tail = 0;
        }

        void UART0_ClearRxBuffer(void)
        {
            uart0_rx_head = 0;
            uart0_rx_tail = 0;
        }

        unsigned char UART0_TxBufferEmpty(void)
        {
            return (uart0_tx_head == uart0_tx_tail);
        }

        unsigned char UART0_RxBufferEmpty(void)
        {
            return (uart0_rx_head == uart0_rx_tail);
        }

        unsigned char UART0_TxBufferFull(void)
        {
            unsigned char next = uart0_tx_head + 1;
            if (next >= UART0_TX_BUFSIZE)
                next = 0;

            return (next == uart0_tx_tail);
        }

        unsigned char UART0_RxBufferFull(void)
        {
            unsigned char next = uart0_rx_head + 1;
            if (next >= UART0_RX_BUFSIZE)
                next = 0;

            return (next == uart0_rx_tail);
        }

#endif //UART0_INT_DRIVEN


    unsigned char UART0_RXready(void)
    {
        #ifdef UART0_INT_DRIVEN
            return (uart0_rx_head != uart0_rx_tail);
        #else
            return ((IFG1 & URXIFG0) != 0);
        #endif
    }

    char UART0_getchar(void)
    {
        #ifdef UART0_INT_DRIVEN
            char c;

            /* wait forever for character to arrive */
            while (uart0_rx_tail == uart0_rx_head);

            c = uart0_rx_buf[uart0_rx_head];
            if (++uart0_rx_head >= UART0_RX_BUFSIZE)
                uart0_rx_head = 0;

			// receive buffer empty
            if (uart0_rx_tail == uart0_rx_head)
			{
            	uart0_rx_head = 0;
            	uart0_rx_tail = 0;
			}

            //since we have incremented the head pointer and not the tail pointer, the
            //buffer is no longer full, therefore we can NEGATE the RTS line
            #ifdef UART0_USING_CTS_RTS
                NEGATE_UART0_RTS();
            #endif // UART0_USING_CTS_RTS

            return c;
        #else
            for (;;)
            {
                if (U0RCTL & RXERR)
                {
                    // if you want to answer receive errors, do it here.
                    U0RCTL &= ~(FE+PE+OE+BRK+RXERR);
                }
                else
                {
                    if (IFG1 & URXIFG0)
                    {  // we've received a character
                        return U0RXBUF;
                    }
                }
            }
        #endif //UART0_INT_DRIVEN
    }

    char UART0_putchar(char c)
    {
        #ifdef UART0_INT_DRIVEN
            unsigned char next;

            next = uart0_tx_head + 1;
            if (next >= UART0_TX_BUFSIZE)
                next = 0;

            /* wait forever until buffer is not full */
            /* apr 25 01 - jcc changed to dump bytes instead of waiting */
            //while (next == uart0_tx_tail);
            if (next == uart0_tx_tail)
                return 0;
            uart0_tx_buf[uart0_tx_head] = c;  // store character in buffer
            uart0_tx_head = next;

            #ifdef UART0_USING_CTS_RTS
                //the transmit buffer is free for us to write to now
                //only write if the CTS line is high
                if (uart0_tx_enable)
                {
            #endif //UART0_USING_CTS_RTS

                    // if not already transmitting, then start transmit interrupt
                    if (!(uart0_flags & UART_TRANSMITTING))
                    {
                        U0TXBUF = uart0_tx_buf[uart0_tx_tail];
                        if (++uart0_tx_tail >= UART0_TX_BUFSIZE)
                        uart0_tx_tail = 0;
                        uart0_flags |= UART_TRANSMITTING;      /* set transmitting flag */
                    }

            #ifdef UART0_USING_CTS_RTS
                //the transmit buffer is free for us to write to now
                //wait for the receiver to set its CTS
                } //end if (uart0_tx_enable)
            #endif //UART0_USING_CTS_RTS

            return c;
        #else
            // wait for other character to transmit
            while (!(IFG1 & UTXIFG0));

            #ifdef UART0_USING_CTS_RTS
                //the transmit buffer is free for us to write to now
                //wait for the receiver to set its CTS
                while (!uart0_tx_enable);
            #endif //UART0_USING_CTS_RTS

            U0TXBUF = c;
            return c;
        #endif //UART0_INT_DRIVEN
    }

    void UART0_putstring(char *s)
    {
        while(*s)
            UART0_putchar(*s++);
    }

#endif // USE_UART0


////////////////////////////////////////////////////////////////
//   UART1 functions
////////////////////////////////////////////////////////////////
#ifdef USE_UART1
    #ifdef UART1_INT_DRIVEN
        void UART1_ClearTxBuffer(void)
        {
            uart1_tx_head = 0;
            uart1_tx_tail = 0;
        }

        void UART1_ClearRxBuffer(void)
        {
            uart1_rx_head = 0;
            uart1_rx_tail = 0;
        }

        void UART1_wait_tx_empty(void)
        {
            while(!UART1_TxBufferEmpty());
            while(!(UTCTL0 & TXEPT));
        }

        unsigned char UART1_TxBufferEmpty(void)
        {
            return (uart1_tx_head == uart1_tx_tail);
        }

        unsigned char UART1_RxBufferEmpty(void)
        {
            return (uart1_rx_head == uart1_rx_tail);
        }

        unsigned char UART1_TxBufferFull(void)
        {
            unsigned char next = uart1_tx_head + 1;
            if (next >= UART1_TX_BUFSIZE)
                next = 0;

            return (next == uart1_tx_tail);
        }

        unsigned char UART1_RxBufferFull(void)
        {
            unsigned char next = uart1_rx_head + 1;
            if (next >= UART1_RX_BUFSIZE)
                next = 0;

            return (next == uart1_rx_tail);
        }
    #endif //UART1_INT_DRIVEN


    unsigned char UART1_RXready(void)
    {
        #ifdef UART1_INT_DRIVEN
            return (uart1_rx_head != uart1_rx_tail);
        #else
            return ((IFG2 & URXIFG1) != 0);
        #endif
    }

    char UART1_getchar(void)
    {
        #ifdef UART1_INT_DRIVEN
            char c;

            /* wait forever for character to arrive */
            while (uart1_rx_tail == uart1_rx_head);

            c = uart1_rx_buf[uart1_rx_head];
            if (++uart1_rx_head >= UART1_RX_BUFSIZE)
                uart1_rx_head = 0;

			// receive buffer empty
            if (uart1_rx_tail == uart1_rx_head)
			{
            	uart1_rx_head = 0;
            	uart1_rx_tail = 0;
			}

            //since we have incremented the head pointer and not the tail pointer, the
            //buffer is no longer full, therefore we can NEGATE the RTS line
            #ifdef UART1_USING_CTS_RTS
                NEGATE_UART1_RTS();
            #endif // UART1_USING_CTS_RTS
            return c;
        #else
            for (;;)
            {
                if (U1RCTL & RXERR)
                {
                    // if you want to answer receive errors, do it here.
                    U1RCTL &= ~(FE+PE+OE+BRK+RXERR);
                }
                else
                {
                    if (IFG2 & URXIFG1)
                    {  // we've received a character
                        return U1RXBUF;
                    }
                }
            }
        #endif //UART1_INT_DRIVEN
    }


    char UART1_putchar(char c)
    {
        #ifdef UART1_INT_DRIVEN
            unsigned char next;

            next = uart1_tx_head + 1;
            if (next >= UART1_TX_BUFSIZE)
                next = 0;

            /* wait forever until buffer is not full */
            while (next == uart1_tx_tail);

            uart1_tx_buf[uart1_tx_head] = c;  // store character in buffer
            uart1_tx_head = next;

            #ifdef UART1_USING_CTS_RTS
                //the transmit buffer is free for us to write to now
                //wait for the receiver to set its CTS
                if (uart1_tx_enable)
                {
            #endif //UART1_USING_CTS_RTS

                    // if not already transmitting, then start transmit interrupt
                    if (!(uart1_flags & UART_TRANSMITTING))
                    {
                        U1TXBUF = uart1_tx_buf[uart1_tx_tail];
                        if (++uart1_tx_tail >= UART1_TX_BUFSIZE)
                        uart1_tx_tail = 0;
                        uart1_flags |= UART_TRANSMITTING;      /* set transmitting flag */
                    }

            #ifdef UART1_USING_CTS_RTS
                //the transmit buffer is free for us to write to now
                //wait for the receiver to set its CTS
                } //end if (uart1_tx_enable)
            #endif //UART1_USING_CTS_RTS

                return c;
        #else
            // wait for other character to transmit
            while (!(IFG2 & UTXIFG1));


            #ifdef UART1_USING_CTS_RTS
                //the transmit buffer is free for us to write to now
                //wait for the receiver to set its CTS
                while (!uart1_tx_enable);
            #endif //UART1_USING_CTS_RTS

            U1TXBUF = c;
            return c;
        #endif //UART1_INT_DRIVEN
    }

    void UART1_putstring(char *s)
    {
        while(*s)
            UART1_putchar(*s++);
    }

#endif /* USE_UART1 */

void UART_RTSCTS_init(void)
{
    ////////////////////////////////////////////////////////////////
    //   CTS/RTS Initialization
    ////////////////////////////////////////////////////////////////
    #ifdef UART0_USING_CTS_RTS
        #ifdef USE_UART0
            CTS_RTS_setup();
        #endif /* USE_UART0 */
    #endif /* UART0_USING_CTS_RTS */

    ////////////////////////////////////////////////////////////////
    //   MAXIM CHIP Initialization
    ////////////////////////////////////////////////////////////////
    #ifdef USING_MAXIM_CE
        P1DIR |= MAXIM_CE_PIN;
        P1SEL &= ~MAXIM_CE_PIN;
        //reverse logic to enable maxim chip; set CE to low
        P1OUT &= ~MAXIM_CE_PIN;
    #endif //USING_MAXIM_CE
}

#ifdef UART0_INT_DRIVEN
    #ifdef USE_UART0
		////////////////////////////////////////////////////////////////
		//   Interrupt Service Routines
		////////////////////////////////////////////////////////////////
		#pragma vector=UART0RX_VECTOR
		__interrupt void USART0_RX_ISR (void)
		{
		//***** RX Interrupt Service Routine
			unsigned char next;
			unsigned char ch;
			
			/* if we received a bad character, replace received data with 0x00. */
			ch = U0RXBUF;
			if (U0RCTL & RXERR)
			{
				U0RCTL &= ~(FE+PE+OE+BRK+RXERR);
				ch = 0x00;
			}
			next = uart0_rx_tail + 1;
			if (next >= UART0_RX_BUFSIZE)
				next = 0;
			if (next != uart0_rx_head)
			{
				uart0_rx_buf[uart0_rx_tail] = ch;
				uart0_rx_tail = next;
				uart0_flags &= ~UART_RX_OVERFLOW;     /* clear RX buffer full flag */
			}
			else
			{
				uart0_flags |= UART_RX_OVERFLOW;      /* set RX buffer full flag */
			}
			if (ch == CR)
			{
				uart0_flags=UART_CR_RECEIVED;
				_BIC_SR_IRQ(LPM0_bits); // Resume processing in the main loop, clear LPM bits
			}
			//TXBUF0 = RXBUF0;                          // RXBUF0 to TXBUF0
			//LEDTOGGLE2;

		} //***** end RX_ISR

		#pragma vector=UART0TX_VECTOR
		__interrupt void USART0_TX_ISR (void)
		//***** TX Interrupt Service Routine
        {
            if (uart0_tx_tail != uart0_tx_head) /* check if buffer is empty */
            {
                #ifdef UART0_USING_CTS_RTS
                //the transmit buffer is free for us to write to now
                //wait for the receiver to set its CTS
                if (uart0_tx_enable)
                {
                #endif //UART0_USING_CTS_RTS

                U0TXBUF = uart0_tx_buf[uart0_tx_tail];
                if (++uart0_tx_tail >= UART0_TX_BUFSIZE)
                uart0_tx_tail = 0;
                uart0_flags |= UART_TRANSMITTING;       /* set transmitting flag */

                #ifdef UART0_USING_CTS_RTS
                //the transmit buffer is free for us to write to now
                //wait for the receiver to set its CTS
                }
                #endif //UART0_USING_CTS_RTS
            }
            else
            {
                uart0_flags &= ~UART_TRANSMITTING;      /* clear transmitting flag */
            }
        }

    #endif //USE_UART0
#endif // UART0_INT_DRIVEN


#ifdef UART0_USING_CTS_RTS //|| UART1_USING_CTS_RTS)
    #ifdef USE_UART0
        /* RS232 CTS_RTS setup routine */
        void CTS_RTS_setup(void)
        {
            /* set up pin usage */
            #ifdef UART0_USING_CTS_RTS
                P1DIR &= ~UART0_CTS_PIN;  /* use the CTS pin as an input   */
                P1DIR |= UART0_RTS_PIN;   /* use the RTS pin as an output  */
                P1SEL &= ~UART0_CTS_PIN & ~UART0_RTS_PIN;
                P1IE |= UART0_CTS_PIN; /* enable interrupts for the CTS pin */
                //check to see if the CTS is currently high or low
                if (P1IN & UART0_CTS_PIN)
                {
                    //since CTS is currently high, we need to look for falling edge
                    SELECT_FALL_EDGE_UART0_CTS;  /* falling edge */
                    /* because our CTS and RTS signals (at least on our side of the MAX
                        * chip) are reverse logic, we need to disable the transmission of
                        * serial data
                        */
                    uart0_tx_enable = false;
                        P1OUT |= BIT1;  // Set P1.0
                        P1OUT &= ~BIT0; // Clear P1.1
                }
                else
                {
                    //since CTS is currently low, we need to look for a rising edge
                        SELECT_RISE_EDGE_UART0_CTS;   /* default - rising edge */
                        /* because our CTS and RTS signals (at least on our side of the MAX
                        * chip) are reverse logic, we need to disable the transmission of
                        * serial data
                        */
                        uart0_tx_enable = true;
                        P1OUT |= BIT0;  // Set P1.0
                        P1OUT &= ~BIT1; // Clear P1.1
                }
                ASSERT_UART0_RTS();
            #endif //UART0_USING_CTS_RTS

            //add section for UART1, this will depend on the pins used
            /* set up pin usage */
            #ifdef UART1_USING_CTS_RTS
                P1SEL &= ~UART1_CTS_PIN;  /* use the CTS pin as an input   */
                P1SEL |= UART1_RTS_PIN;   /* use the RTS pin as an output  */
                P1IE = UART1_CTS_PIN; /* enable interrupts for the CTS pin */
                //check to see if the CTS is currently high or low
                if (P1IN & UART1_CTS_PIN)
                {
                    //since CTS is currently high, we need to look for falling edge
                    SELECT_FALL_EDGE_UART1_CTS;  /* falling edge */
                }
                else
                {
                    //since CTS is currently low, we need to look for a rising edge
                    SELECT_FALL_EDGE_UART1_CTS;   /* default - rising edge */
                }
            register_isr(PORT1_VECTOR, PORT_ISR);
            #endif //UART1_USING_CTS_RTS
        }

//interrupt routine for the incoming CTS line
        /***********************************************************************/
        /* External port interrupt                                             */
        /***********************************************************************/
        void PORT_ISR( void )
        {
            #ifdef UART0_USING_CTS_RTS
                /* check to see if an interrupt has occurred for uart 0 cts          */
                if (P1IFG & UART0_CTS_INT)   /* if bit for uart0 cts interrupt is set*/
                {

                    if (P1IN & UART0_CTS_PIN) /* if we were waiting for rising edge   */
                    {
                        /* set the interrupt to look for a falling edge                */
                        SELECT_FALL_EDGE_UART0_CTS;
                        /* because our CTS and RTS signals (at least on our side of the MAX
                        * chip) are reverse logic, we need to disable the transmission of
                        * serial data
                        */
                        uart0_tx_enable = false;
                        P1OUT |= BIT1;  // Set P1.0
                        P1OUT &= ~BIT0; // Clear P1.1


                    }
                    else  /* if we were waiting for falling edge                      */
                    {
                        /* set the interrupt to look for a rising edge                 */
                        SELECT_RISE_EDGE_UART0_CTS;
                        /* because our CTS and RTS signals (at least on our side of the MAX
                        * chip) are reverse logic, we need to disable the transmission of
                        * serial data
                        */
                        uart0_tx_enable = true;
                        P1OUT |= BIT0;  // Set P1.0
                        P1OUT &= ~BIT1; // Clear P1.1

                        //since we just received the CTS signal
                    //check to see if we have data to transmit
                    if (uart0_tx_tail != uart0_tx_head) /* check if buffer is empty */
                    {
                        U0TXBUF = uart0_tx_buf[uart0_tx_tail];
                        if (++uart0_tx_tail >= UART0_TX_BUFSIZE)
                            uart0_tx_tail = 0;
                            uart0_flags |= UART_TRANSMITTING;       /* set transmitting flag */
                        }
                        else //if (uart0_tx_tail...
                        {
                            uart0_flags &= ~UART_TRANSMITTING;      /* clear transmitting flag */
                        } //end if (uart0_tx_tail...
                    } //end if (P1IN & UART0_CTS_PIN)
                    /* reset interrupt flag for uart0 cts                             */
                    P1IFG &= ~UART0_CTS_INT;
                }
            #endif //UART0_USING_CTS_RTS

            #if 0
                //this needs to be updated to allow use of CTS/RTS on UART1
                #ifdef UART1_USING_CTS_RTS
                    /* check to see if an interrupt has occurred for uart1 cts           */
                    if (P1IFG & UART1_CTS_INT)   /* if bit for uart1 cts interrupt is set*/
                    {
                        if (P1IN & UART1_CTS_PIN) /* if we were waiting for rising edge   */
                        {
                            /* set the interrupt to look for a failing edge                */
                            SELECT_FALL_EDGE_UART1_CTS;
                            uart1_tx_enable = true;

                            //since we just received the CTS signal
                            //check to see if we have data to transmit
                            if (uart1_tx_tail != uart1_tx_head) /* check if buffer is empty */
                            {
                                U1TXBUF = uart1_tx_buf[uart1_tx_tail];
                                if (++uart1_tx_tail >= UART1_TX_BUFSIZE)
                                    uart1_tx_tail = 0;
                                uart1_flags |= UART_TRANSMITTING;       /* set transmitting flag */
                            }
                            else
                            {
                                uart1_flags &= ~UART_TRANSMITTING;      /* clear transmitting flag */
                            }
                        }
                        else  /* if we were waiting for failing edge                      */
                        {
                            /* set the interrupt to look for a rising edge                 */
                            SELECT_RISE_EDGE_UART1_CTS;
                            uart1_tx_enable = false;
                        }
                        /* reset interrupt flag for uart1 cts                             */
                        P1IFG &= ~UART1_CTS_INT;
                    }
                #endif //UART1_USING_CTS_RTS
            #endif //0
        }
    #endif /* USE_UART0 */
#endif /* UART0_USING_CTS_RTS || UART1_USING_CTS_RTS */

